const path = require("path");
const webpack = require("webpack");
const progress = require("progress");
const chalk = require("chalk");
const HtmlWebpackPlugin = require("html-webpack-plugin");
const _ = require("lodash");
const autoEntry = require("./autoEntry");

const os = require("os");

const cpuCount = os.cpus().length;

const cliNodeModules = path.resolve(__dirname, "../../node_modules");
const cliRoot = path.resolve(__dirname);

//min
const UglifyJSPlugin = require("uglifyjs-webpack-plugin");

const OptimizeCSSAssetsPlugin = require("optimize-css-assets-webpack-plugin");

const defaultPlugins = require("./defaultPlugins");

const miniCssLoaders = require("./miniCssLoaders");

const assetsLoader = require("./assetsLoader");

const cssnanoConf = require("./cssnanoConf");

module.exports = function(program, ksr, isDev) {
  const entry = {};
  const pathConvert = ksr.pathConvert;
  const min = program.min === true;
  const entrys = autoEntry(program.entry, ksr);
  const plugins = defaultPlugins(program, ksr, isDev);
  const filename = ksr.filename;
  const chunkFilename = ksr.chunkFilename;

  const assetsPublicPath = ksr.assetsPublicPath;

  const srcPath = pathConvert(ksr.src);
  const rootPath = pathConvert(ksr.root);
  const distPath = pathConvert(ksr.dist);
  const nodeModulesPath = pathConvert("node_modules");

  const minifyHtml = {
    removeComments: true,
    collapseWhitespace: true,
    collapseBooleanAttributes: true,
    removeRedundantAttributes: true,
    removeEmptyAttributes: true,
    caseSensitive: true,
    ignoreCustomFragments: [/<%[\s\S]*?%>/, /<\?[\s\S]*?\?>/, /\{\{.+?\}\}/]
  };
  _.each(entrys, function(_entry) {
    if (_entry["template"]) {
      plugins.push(
        new HtmlWebpackPlugin({
          cache: true,
          showErrors: true,
          chunks: ["__vendors__", _entry.chunk],
          minify: min !== true ? false : minifyHtml,
          filename: _entry.filename,
          template: _entry.template
        })
      );
    }
    entry[_entry.chunk] = [];
    if (_entry.js) {
      entry[_entry.chunk].push(_entry.js);
    }
    if (_entry.ts) {
      entry[_entry.chunk].push(_entry.ts);
    }
  });

  const webpackConfig = {
    mode: "none",
    entry: entry,
    cache: isDev,
    output: {
      path: distPath,
      filename: `${filename}.js`,
      chunkFilename: `${chunkFilename}.js`,
      publicPath: assetsPublicPath,
      pathinfo: isDev
    },
    devtool: isDev ? "cheap-module-eval-source-map" : "source-map",
    resolveLoader: {
      modules: [
        nodeModulesPath,
        "node_modules",
        cliNodeModules,
        rootPath,
        srcPath,
        cliRoot
      ]
    },
    performance: {
      maxAssetSize: 3072000, // 3Mib
      maxEntrypointSize: 3072000 // 3Mib
    },
    optimization: {
      //优化构建项 , 开发 与 构建时 分开处理
      noEmitOnErrors: true, //编译时出现错误时，使用跳过发射阶段
      runtimeChunk: false, //入口点添加一个额外的块

      // 开发时启用
      namedModules: isDev, //可读模块标识符以获得更好的调试
      namedChunks: isDev, //可读模块标识符以获得更好的调试
      // 开发时启用

      // 构建启用
      removeAvailableModules: !isDev, //移除未使用快
      removeEmptyChunks: !isDev, //移除空快
      mergeDuplicateChunks: !isDev, // 合并重复快

      usedExports: !isDev,
      concatenateModules: !isDev,
      occurrenceOrder: !isDev,
      // sideEffects: !isDev, //tree shaking 会删除文件  还是不用这个了
      flagIncludedChunks: !isDev, //webpack确定并标记作为其他块的子集的块，其方式是当已经加载较大的块时不必加载子集
      // 构建启用

      minimize: min, //webpack使用UglifyjsWebpackPlugin最小化捆绑包
      minimizer: min
        ? [
            new UglifyJSPlugin({
              cache: true,
              parallel: cpuCount - 1,
              uglifyOptions: {
                warnings: false,
                output: {
                  comments: false
                },
                compress: {
                  drop_console: true
                },
                dead_code: true
              }
            }),
            new OptimizeCSSAssetsPlugin({
              cssProcessorOptions: cssnanoConf
            })
          ]
        : [],
      splitChunks: {
        chunks: "all",
        minSize: 1000,
        minChunks: 2,
        maxAsyncRequests: 5,
        maxInitialRequests: 3,
        name: true,
        cacheGroups: {
          default: {
            name: "__vendors__",
            chunks: "all",
            reuseExistingChunk: true
          }
        }
      }
    },
    resolve: {
      // 用于查找模块的目录
      modules: [
        nodeModulesPath,
        rootPath,
        srcPath,
        cliNodeModules,
        cliRoot,
        "node_modules"
      ],
      extensions: [".js", ".pug", ".json"],
      alias: {
        "~": srcPath,
        "@": rootPath,
        kpc: "kpc/@stylus"
      }
    },
    module: {
      rules: [
        {
          test: /\.m?js$/,
          exclude: [/node_modules/, /bower_components/],
          use: "happypack/loader?id=babel"
        },
        //happypack 不支持 mini-css-extract-plugin loader
        {
          test: /\.[cs]ss$/,
          use: miniCssLoaders("happypack/loader?id=css", isDev)
        },
        {
          test: /\.sass$/,
          use: miniCssLoaders("happypack/loader?id=sass", isDev)
        },
        {
          test: /\.scss$/,
          use: miniCssLoaders("happypack/loader?id=scss", isDev)
        },
        {
          test: /\.styl(us)?$/,
          use: miniCssLoaders("happypack/loader?id=stylus", isDev)
        },
        {
          test: /\.less$/,
          use: miniCssLoaders("happypack/loader?id=less", isDev)
        },
        {
          test: /\.(pug|jade)$/,
          oneOf: [
            // this applies to `<template lang="pug">` in Vue components
            {
              resourceQuery: /^\?vue/,
              use: "happypack/loader?id=vuepug"
            },
            // this applies to pug imports inside JavaScript
            {
              use: "happypack/loader?id=pug"
            }
          ]
        },
        //happypack 会造成编码问题 暂时忽略
        {
          test: /\.(mp4|webm|ogg|mp3|wav|flac|aac|png|jpg|gif|svg|eot|woff2?|ttf|otf|swf)(\?.*)?$/,
          use: assetsLoader(ksr)
        }
      ]
    },
    plugins: plugins
  };
  return webpackConfig;
};
